/*
** prompter.c -- simple prompting editor front-end
**
** This code is Copyright (c) 2002, by the authors of nmh.  See the
** COPYRIGHT file in the root directory of the nmh distribution for
** complete copyright information.
*/

#include <h/mh.h>
#include <fcntl.h>
#include <h/signals.h>
#include <errno.h>
#include <signal.h>
#include <setjmp.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/stat.h>
#include <locale.h>

static struct swit switches[] = {
#define PREPSW  0
	{ "prepend", 0 },
#define NPREPSW  1
	{ "noprepend", 2 },
#define RAPDSW  2
	{ "rapid", 0 },
#define NRAPDSW  3
	{ "norapid", 2 },
#define BODYSW  4
	{ "body", 0 },
#define NBODYSW  5
	{ "nobody", 2 },
#define VERSIONSW 6
	{ "Version", 0 },
#define HELPSW  7
	{ "help", 0 },
	{ NULL, 0 }
};


volatile sig_atomic_t wtuser = 0;
volatile sig_atomic_t sigint = 0;
static jmp_buf sigenv;

/*
** prototypes
*/
int getln(char *, int);
static void intrser(int);


int
main(int argc, char **argv)
{
	int qbody = 1, prepend = 1, rapid = 0;
	int fdi, fdo, i, state;
	char *cp, *drft = NULL;
	char name[NAMESZ], field[BUFSIZ];
	char buffer[BUFSIZ], tmpfil[BUFSIZ];
	char **arguments, **argp;
	FILE *in, *out;
	char *tfile = NULL;

	setlocale(LC_ALL, "");
	invo_name = mhbasename(argv[0]);

	/* read user profile/context */
	context_read();

	arguments = getarguments(invo_name, argc, argv, 1);
	argp = arguments;

	while ((cp = *argp++)) {
		if (*cp == '-') {
			switch (smatch(++cp, switches)) {
			case AMBIGSW:
				ambigsw(cp, switches);
				exit(1);
			case UNKWNSW:
				adios(NULL, "-%s unknown", cp);

			case HELPSW:
				snprintf(buffer, sizeof(buffer),
						"%s [switches] file",
						invo_name);
				print_help(buffer, switches, 1);
				exit(0);
			case VERSIONSW:
				print_version(invo_name);
				exit(0);

			case PREPSW:
				prepend++;
				continue;
			case NPREPSW:
				prepend = 0;
				continue;

			case RAPDSW:
				rapid++;
				continue;
			case NRAPDSW:
				rapid = 0;
				continue;

			case BODYSW:
				qbody++;
				continue;
			case NBODYSW:
				qbody = 0;
				continue;
			}
		} else if (!drft) {
			drft = cp;
		}
	}

	if (!drft)
		adios(NULL, "usage: %s [switches] file", invo_name);
	if ((in = fopen(drft, "r")) == NULL)
		adios(drft, "unable to open");

	tfile = m_mktemp2(NULL, invo_name, NULL, &out);
	if (tfile == NULL)
		adios("prompter", "unable to create temporary file");
	chmod(tmpfil, 0600);
	strncpy(tmpfil, tfile, sizeof(tmpfil));

	sigint = 0;
	SIGNAL2(SIGINT, intrser);

	/*
	** Loop through the lines of the draft skeleton.
	*/
	for (state = FLD;;) {
		switch (state = m_getfld(state, name, field, sizeof(field),
				in)) {
		case FLD:
		case FLDEOF:
		case FLDPLUS:
			/*
			** Check if the value of field contains
			** anything other than space or tab.
			*/
			for (cp = field; *cp; cp++)
				if (*cp != ' ' && *cp != '\t')
					break;

			/* If so, just add header line to draft */
			if (*cp++ != '\n' || *cp) {
				printf("%s:%s", name, field);
				fprintf(out, "%s:%s", name, field);
				while (state == FLDPLUS) {
					state = m_getfld(state, name, field,
							sizeof(field), in);
					printf("%s", field);
					fprintf(out, "%s", field);
				}
			} else {
				/* Else, get value of header field */
				printf("%s: ", name);
				fflush(stdout);
				i = getln(field, sizeof(field));
				if (i == -1) {
abort:
					unlink(tmpfil);
					/* sysexits.h EX_DATAERR */
					exit(1);
				}
				if (i || (field[0]!='\n' && field[0]!='\0')) {
					fprintf(out, "%s:", name);
					do {
						if (field[0] != ' ' && field[0] != '\t')
							putc(' ', out);
						fprintf(out, "%s", field);
					} while (i == 1 && (i = getln(field, sizeof(field))) >= 0);
					if (i == -1)
						goto abort;
				}
			}

			if (state == FLDEOF) {  /* moby hack */
				/* draft has no body separator; only headers */
				fprintf(out, "--------\n");
				if (!qbody)
					break;
				printf("--------\n");
				goto has_no_body;
			}
			continue;

		case BODY:
		case BODYEOF:
		case FILEEOF:
			fprintf(out, "--------\n");
			if (qbody) {
				if (!*field) {
					printf("--------\n");
					goto has_no_body;
				}

				if (prepend) {
					printf("--------Enter initial text\n");
					fflush(stdout);
					for (;;) {
						getln(buffer, sizeof(buffer));
						if (!*buffer)
							break;
						fprintf(out, "%s", buffer);
					}
				} else {
					printf("--------\n");
				}
			}

			do {
				fprintf(out, "%s", field);
				if (!rapid && !sigint)
					printf("%s", field);
			} while (state == BODY &&
					(state = m_getfld(state, name,
					field, sizeof(field), in)));

			if (prepend || !qbody)
				break;

			printf("--------Enter additional text\n");
has_no_body:
			fflush(stdout);
			for (;;) {
				getln(field, sizeof(field));
				if (!*field)
					break;
				fprintf(out, "%s", field);
			}
			break;

		default:
			adios(NULL, "skeleton is poorly formatted");
		}
		break;
	}

	if (qbody)
		printf("--------\n");

	fflush(stdout);
	fclose(in);
	fclose(out);
	SIGNAL(SIGINT, SIG_IGN);

	if ((fdi = open(tmpfil, O_RDONLY)) == NOTOK)
		adios(tmpfil, "unable to re-open");
	if ((fdo = creat(drft, m_gmprot())) == NOTOK)
		adios(drft, "unable to write");
	cpydata(fdi, fdo, tmpfil, drft);
	close(fdi);
	close(fdo);
	unlink(tmpfil);

	context_save();  /* save the context file */
	return 0;
}


int
getln(char *buffer, int n)
{
	int c;
	sig_atomic_t psigint;
	char *cp;

	cp = buffer;
	*cp = '\0';

	switch (setjmp(sigenv)) {
	case 0:
		wtuser = 1;
		psigint = sigint;
		break;

	default:
		wtuser = 0;
		if (sigint == psigint) {
			return 0;
		} else {
			sigint = psigint;
			return NOTOK;
		}
	}

	for (;;) {
		switch (c = getchar()) {
		case EOF:
			clearerr(stdin);
			longjmp(sigenv, DONE);

		case '\n':
			if (cp[-1] == '\\') {
				cp[-1] = c;
				wtuser = 0;
				return 1;
			}
			*cp++ = c;
			*cp = '\0';
			wtuser = 0;
			return 0;

		default:
			if (cp < buffer + n)
				*cp++ = c;
			*cp = '\0';
		}
	}
}


static void
intrser(int i)
{
	if (wtuser) {
		close(STDIN_FILENO);
	}
	sigint++;
}
